@@ -54,6 +54,23 @@ static void        yelp_view_set_property         (GObject            *object,
                                                    const GValue       *value,
                                                    GParamSpec         *pspec);
 
+static void        view_scrolled                  (GtkAdjustment      *adjustment,
+                                                   YelpView           *view);
+static void        view_scroll_adjustments        (YelpView           *view,
+                                                   GtkAdjustment      *hadj,
+                                                   GtkAdjustment      *vadj,
+                                                   gpointer            data);
+static void        popup_open_link                (GtkMenuItem        *item,
+                                                   YelpView           *view);
+static void        popup_open_link_new            (GtkMenuItem        *item,
+                                                   YelpView           *view);
+static void        popup_save_image               (GtkMenuItem        *item,
+                                                   YelpView           *view);
+static void        popup_send_image               (GtkMenuItem        *item,
+                                                   YelpView           *view);
+static void        view_populate_popup            (YelpView           *view,
+                                                   GtkMenu            *menu,
+                                                   gpointer            data);
 static gboolean    view_navigation_requested      (WebKitWebView             *view,
                                                    WebKitWebFrame            *frame,
                                                    WebKitNetworkRequest      *request,
@@ -67,6 +84,8 @@ static void        view_resource_request          (WebKitWebView             *vi
                                                    WebKitNetworkResponse     *response,
                                                    gpointer                   user_data);
 
+static void        view_print                     (GtkAction          *action,
+                                                   YelpView           *view);
 static void        view_history_action            (GtkAction          *action,
                                                    YelpView           *view);
 static void        view_navigation_action         (GtkAction          *action,
@@ -88,6 +107,11 @@ static void        document_callback              (YelpDocument       *document,
                                                    GError             *error);
 
 static const GtkActionEntry entries[] = {
+    {"YelpViewPrint", GTK_STOCK_PRINT,
+     N_("_Print..."),
+     "<Control>P",
+     NULL,
+     G_CALLBACK (view_print) },
     {"YelpViewGoBack", GTK_STOCK_GO_BACK,
      N_("_Back"),
      "<Alt>Left",
@@ -110,6 +134,8 @@ static const GtkActionEntry entries[] = {
      G_CALLBACK (view_navigation_action) }
 };
 
+static gchar *nautilus_sendto = NULL;
+
 enum {
     PROP_0,
     PROP_URI,
@@ -137,8 +163,10 @@ static WebKitWebSettings *websettings;
 typedef struct _YelpBackEntry YelpBackEntry;
 struct _YelpBackEntry {
     YelpUri *uri;
-    gchar *title;
-    gchar *desc;
+    gchar   *title;
+    gchar   *desc;
+    gdouble  hadj;
+    gdouble  vadj;
 };
 static void
 back_entry_free (YelpBackEntry *back)
@@ -158,6 +186,15 @@ struct _YelpViewPrivate {
     gchar         *bogus_uri;
     YelpDocument  *document;
     GCancellable  *cancellable;
+    GtkAdjustment *vadjustment;
+    GtkAdjustment *hadjustment;
+    gdouble        vadjust;
+    gdouble        hadjust;
+    gulong         vadjuster;
+    gulong         hadjuster;
+
+    gchar         *popup_link_uri;
+    gchar         *popup_image_uri;
 
     YelpViewState  state;
 
@@ -197,6 +234,10 @@ yelp_view_init (YelpView *view)
                           G_CALLBACK (view_navigation_requested), NULL);
     g_signal_connect (view, "resource-request-starting",
                       G_CALLBACK (view_resource_request), NULL);
+    g_signal_connect (view, "set-scroll-adjustments",
+                      G_CALLBACK (view_scroll_adjustments), NULL);
+    g_signal_connect (view, "populate-popup",
+                      G_CALLBACK (view_populate_popup), NULL);
 
     priv->action_group = gtk_action_group_new ("YelpView");
     gtk_action_group_set_translation_domain (priv->action_group, GETTEXT_PACKAGE);
@@ -215,6 +256,16 @@ yelp_view_dispose (GObject *object)
         priv->uri = NULL;
     }
 
+    if (priv->vadjuster > 0) {
+        g_source_remove (priv->vadjuster);
+        priv->vadjuster = 0;
+    }
+
+    if (priv->hadjuster > 0) {
+        g_source_remove (priv->hadjuster);
+        priv->hadjuster = 0;
+    }
+
     if (priv->cancellable) {
         g_cancellable_cancel (priv->cancellable);
         g_object_unref (priv->cancellable);
@@ -231,6 +282,7 @@ yelp_view_dispose (GObject *object)
         priv->document = NULL;
     }
 
+    priv->back_cur = NULL;
     while (priv->back_list) {
         back_entry_free ((YelpBackEntry *) priv->back_list->data);
         priv->back_list = g_list_delete_link (priv->back_list, priv->back_list);
@@ -244,6 +296,9 @@ yelp_view_finalize (GObject *object)
 {
     YelpViewPrivate *priv = GET_PRIV (object);
 
+    g_free (priv->popup_link_uri);
+    g_free (priv->popup_image_uri);
+
     g_free (priv->page_id);
     g_free (priv->root_title);
     g_free (priv->page_title);
@@ -261,6 +316,8 @@ yelp_view_class_init (YelpViewClass *klass)
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
     YelpSettings *settings = yelp_settings_get_default ();
 
+    nautilus_sendto = g_find_program_in_path ("nautilus-sendto");
+
     websettings = webkit_web_settings_new ();
     g_signal_connect (settings,
                       "fonts-changed",
@@ -283,8 +340,8 @@ yelp_view_class_init (YelpViewClass *klass)
 		      G_TYPE_FROM_CLASS (klass),
 		      G_SIGNAL_RUN_LAST,
                       0, NULL, NULL,
-		      g_cclosure_marshal_VOID__STRING,
-		      G_TYPE_NONE, 1, G_TYPE_STRING);
+		      g_cclosure_marshal_VOID__OBJECT,
+		      G_TYPE_NONE, 1, YELP_TYPE_URI);
 
     signals[EXTERNAL_URI] =
 	g_signal_new ("external-uri",
@@ -327,7 +384,7 @@ yelp_view_class_init (YelpViewClass *klass)
                                      PROP_PAGE_ID,
                                      g_param_spec_string ("page-id",
                                                           N_("Page ID"),
-                                                          N_("The ID of the root page of the page being viewew"),
+                                                          N_("The ID of the root page of the page being viewed"),
                                                           NULL,
                                                           G_PARAM_READABLE |
                                                           G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
@@ -336,7 +393,7 @@ yelp_view_class_init (YelpViewClass *klass)
                                      PROP_ROOT_TITLE,
                                      g_param_spec_string ("root-title",
                                                           N_("Root Title"),
-                                                          N_("The title of the root page of the page being viewew"),
+                                                          N_("The title of the root page of the page being viewed"),
                                                           NULL,
                                                           G_PARAM_READABLE |
                                                           G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
@@ -395,7 +452,10 @@ yelp_view_get_property (GObject    *object,
             g_value_set_string (value, priv->page_desc);
             break;
         case PROP_PAGE_ICON:
-            g_value_set_string (value, priv->page_icon);
+            if (priv->page_icon)
+                g_value_set_string (value, priv->page_icon);
+            else
+                g_value_set_string (value, "help-contents");
             break;
         case PROP_STATE:
             g_value_set_enum (value, priv->state);
@@ -453,6 +513,7 @@ yelp_view_load_uri (YelpView *view,
                     YelpUri  *uri)
 {
     YelpViewPrivate *priv = GET_PRIV (view);
+    GParamSpec *spec;
 
     view_clear_load (view);
     g_object_set (view, "state", YELP_VIEW_STATE_LOADING, NULL);
@@ -467,11 +528,26 @@ yelp_view_load_uri (YelpView *view,
     priv->page_title = NULL;
     priv->page_desc = NULL;
     priv->page_icon = NULL;
-    g_signal_emit_by_name (view, "notify::page-id", 0);
-    g_signal_emit_by_name (view, "notify::root-title", 0);
-    g_signal_emit_by_name (view, "notify::page-title", 0);
-    g_signal_emit_by_name (view, "notify::page-desc", 0);
-    g_signal_emit_by_name (view, "notify::page-icon", 0);
+
+    spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                         "page-id");
+    g_signal_emit_by_name (view, "notify::page-id", spec);
+
+    spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                         "root-title");
+    g_signal_emit_by_name (view, "notify::root-title", spec);
+
+    spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                         "page-title");
+    g_signal_emit_by_name (view, "notify::page-title", spec);
+
+    spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                         "page-desc");
+    g_signal_emit_by_name (view, "notify::page-desc", spec);
+
+    spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                         "page-icon");
+    g_signal_emit_by_name (view, "notify::page-icon", spec);
 
     gtk_action_set_sensitive (gtk_action_group_get_action (priv->action_group,
                                                            "YelpViewGoPrevious"),
@@ -529,6 +605,302 @@ yelp_view_get_action_group (YelpView *view)
 
 /******************************************************************************/
 
+static void
+view_scrolled (GtkAdjustment *adjustment,
+               YelpView      *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    if (priv->back_cur == NULL || priv->back_cur->data == NULL)
+        return;
+    if (adjustment == priv->vadjustment)
+        ((YelpBackEntry *) priv->back_cur->data)->vadj = gtk_adjustment_get_value (adjustment);
+    else if (adjustment = priv->hadjustment)
+        ((YelpBackEntry *) priv->back_cur->data)->hadj = gtk_adjustment_get_value (adjustment);
+}
+
+static void
+view_scroll_adjustments (YelpView      *view,
+                         GtkAdjustment *hadj,
+                         GtkAdjustment *vadj,
+                         gpointer       data)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    priv->vadjustment = vadj;
+    if (priv->vadjuster > 0)
+        g_source_remove (priv->vadjuster);
+    priv->vadjuster = 0;
+    if (vadj) {
+        priv->vadjuster = g_signal_connect (vadj, "value-changed",
+                                            G_CALLBACK (view_scrolled), view);
+    }
+    priv->hadjustment = hadj;
+    if (priv->hadjuster > 0)
+        g_source_remove (priv->hadjuster);
+    priv->hadjuster = 0;
+    if (hadj) {
+        priv->hadjuster = g_signal_connect (hadj, "value-changed",
+                                            G_CALLBACK (view_scrolled), view);
+    }
+}
+
+static void
+popup_open_link (GtkMenuItem *item,
+                 YelpView    *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    YelpUri *uri;
+
+    if (g_str_has_prefix (priv->popup_link_uri, BOGUS_URI))
+        uri = yelp_uri_new_relative (priv->uri, priv->popup_link_uri + BOGUS_URI_LEN);
+    else
+        uri = yelp_uri_new_relative (priv->uri, priv->popup_link_uri);
+
+    yelp_view_load_uri (view, uri);
+
+    g_free (priv->popup_link_uri);
+    priv->popup_link_uri = NULL;
+}
+
+static void
+popup_open_link_new (GtkMenuItem *item,
+                     YelpView    *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    YelpUri *uri;
+
+    if (g_str_has_prefix (priv->popup_link_uri, BOGUS_URI))
+        uri = yelp_uri_new_relative (priv->uri, priv->popup_link_uri + BOGUS_URI_LEN);
+    else
+        uri = yelp_uri_new_relative (priv->uri, priv->popup_link_uri);
+
+    g_free (priv->popup_link_uri);
+    priv->popup_link_uri = NULL;
+
+    g_signal_emit (view, signals[NEW_VIEW_REQUESTED], 0, uri);
+    g_object_unref (uri);
+}
+
+typedef struct _YelpSaveData YelpSaveData;
+struct _YelpSaveData {
+    GFile     *orig;
+    GFile     *dest;
+    YelpView  *view;
+    GtkWindow *window;
+};
+
+static void
+file_copied (GFile        *file,
+             GAsyncResult *res,
+             YelpSaveData *data)
+{
+    GError *error = NULL;
+    if (!g_file_copy_finish (file, res, &error)) {
+        GtkWidget *dialog = gtk_message_dialog_new (GTK_WIDGET_VISIBLE (data->window) ? data->window : NULL,
+                                                    GTK_DIALOG_DESTROY_WITH_PARENT,
+                                                    GTK_MESSAGE_ERROR,
+                                                    GTK_BUTTONS_OK,
+                                                    "%s", error->message);
+        gtk_dialog_run (GTK_DIALOG (dialog));
+        gtk_widget_destroy (dialog);
+    }
+    g_object_unref (data->orig);
+    g_object_unref (data->dest);
+    g_object_unref (data->view);
+    g_object_unref (data->window);
+}
+
+static void
+popup_save_image (GtkMenuItem *item,
+                  YelpView    *view)
+{
+    YelpSaveData *data;
+    GtkWidget *dialog, *window;
+    gchar *basename;
+    gint res;
+    YelpViewPrivate *priv = GET_PRIV (view);
+
+    for (window = gtk_widget_get_parent (GTK_WIDGET (view));
+         window && !GTK_IS_WINDOW (window);
+         window = gtk_widget_get_parent (window));
+
+    data = g_new0 (YelpSaveData, 1);
+    data->orig = g_file_new_for_uri (priv->popup_image_uri);
+    data->view = g_object_ref (view);
+    data->window = g_object_ref (window);
+    g_free (priv->popup_image_uri);
+    priv->popup_image_uri = NULL;
+
+    dialog = gtk_file_chooser_dialog_new (_("Save Image"),
+                                          GTK_WINDOW (window),
+                                          GTK_FILE_CHOOSER_ACTION_SAVE,
+                                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+                                          GTK_STOCK_SAVE, GTK_RESPONSE_OK,
+                                          NULL);
+    gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
+    basename = g_file_get_basename (data->orig);
+    gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), basename);
+    g_free (basename);
+    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog),
+                                         g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP));
+
+    res = gtk_dialog_run (GTK_DIALOG (dialog));
+
+    if (res == GTK_RESPONSE_OK) {
+        data->dest = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (dialog));
+        g_file_copy_async (data->orig, data->dest,
+                           G_FILE_COPY_OVERWRITE,
+                           G_PRIORITY_DEFAULT,
+                           NULL, NULL, NULL,
+                           (GAsyncReadyCallback) file_copied,
+                           data);
+    }
+    else {
+        g_object_unref (data->orig);
+        g_object_unref (data->view);
+        g_object_unref (data->window);
+        g_free (data);
+    }
+
+    gtk_widget_destroy (dialog);
+}
+
+static void
+popup_send_image (GtkMenuItem *item,
+                  YelpView    *view)
+{
+    gchar *command;
+    YelpViewPrivate *priv = GET_PRIV (view);
+
+    command = g_strdup_printf ("%s %s", nautilus_sendto, priv->popup_image_uri);
+
+    gdk_spawn_command_line_on_screen (gtk_widget_get_screen (GTK_WIDGET (view)),
+                                      command, NULL);
+
+    g_free (command);
+    g_free (priv->popup_image_uri);
+    priv->popup_image_uri = NULL;
+}
+
+static void
+view_populate_popup (YelpView *view,
+                     GtkMenu  *menu,
+                     gpointer  data)
+{
+    WebKitHitTestResult *result;
+    WebKitHitTestResultContext context;
+    GdkEvent *event;
+    YelpViewPrivate *priv = GET_PRIV (view);
+    GList *children;
+    GtkWidget *item;
+
+    children = gtk_container_get_children (GTK_CONTAINER (menu));
+    while (children) {
+        gtk_container_remove (GTK_CONTAINER (menu),
+                              GTK_WIDGET (children->data));
+        children = children->next;
+    }
+    g_list_free (children);
+
+    event = gtk_get_current_event ();
+
+    result = webkit_web_view_get_hit_test_result (WEBKIT_WEB_VIEW (view), (GdkEventButton *) event);
+    g_object_get (result, "context", &context, NULL);
+
+    if (context & WEBKIT_HIT_TEST_RESULT_CONTEXT_LINK) {
+        gchar *uri;
+        g_object_get (result, "link-uri", &uri, NULL);
+        g_free (priv->popup_link_uri);
+        priv->popup_link_uri = uri;
+
+        if (g_str_has_prefix (priv->popup_link_uri, "mailto:")) {
+            /* Not using a mnemonic because underscores are common in email
+             * addresses, and we'd have to escape them. There doesn't seem
+             * to be a quick GTK+ function for this. In practice, there will
+             * probably only be one menu item for mailto link popups anyway,
+             * so the mnemonic's not that big of a deal.
+             */
+            gchar *label = g_strdup_printf (_("Send email to %s"),
+                                            priv->popup_link_uri + 7);
+            item = gtk_menu_item_new_with_label (label);
+            g_signal_connect (item, "activate",
+                              G_CALLBACK (popup_open_link), view);
+            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+        }
+        else {
+            item = gtk_menu_item_new_with_mnemonic (_("_Open Link"));
+            g_signal_connect (item, "activate",
+                              G_CALLBACK (popup_open_link), view);
+            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+            item = gtk_menu_item_new_with_mnemonic (_("Open Link in New _Window"));
+            g_signal_connect (item, "activate",
+                              G_CALLBACK (popup_open_link_new), view);
+            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+        }
+    }
+    else {
+        item = gtk_action_create_menu_item (gtk_action_group_get_action (priv->action_group,
+                                                                         "YelpViewGoBack"));
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+        item = gtk_action_create_menu_item (gtk_action_group_get_action (priv->action_group,
+                                                                         "YelpViewGoForward"));
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+    }
+
+    if ((context & WEBKIT_HIT_TEST_RESULT_CONTEXT_IMAGE) ||
+        (context & WEBKIT_HIT_TEST_RESULT_CONTEXT_MEDIA)) {
+        /* This doesn't currently work for video with automatic controls,
+         * because WebKit puts the hit test on the div with the controls.
+         */
+        gboolean image = context & WEBKIT_HIT_TEST_RESULT_CONTEXT_IMAGE;
+        gchar *uri;
+        g_object_get (result, image ? "image-uri" : "media-uri", &uri, NULL);
+        g_free (priv->popup_image_uri);
+        if (g_str_has_prefix (uri, BOGUS_URI)) {
+            priv->popup_image_uri = yelp_uri_locate_file_uri (priv->uri, uri + BOGUS_URI_LEN);
+            g_free (uri);
+        }
+        else {
+            priv->popup_image_uri = uri;
+        }
+
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        if (image)
+            item = gtk_menu_item_new_with_mnemonic (_("_Save Image As..."));
+        else
+            item = gtk_menu_item_new_with_mnemonic (_("_Save Video As..."));
+        g_signal_connect (item, "activate",
+                          G_CALLBACK (popup_save_image), view);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        if (nautilus_sendto) {
+            if (image)
+                item = gtk_menu_item_new_with_mnemonic (_("S_end Image To..."));
+            else
+                item = gtk_menu_item_new_with_mnemonic (_("S_end Video To..."));
+            g_signal_connect (item, "activate",
+                              G_CALLBACK (popup_send_image), view);
+            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+        }
+    }
+
+    if (context & WEBKIT_HIT_TEST_RESULT_CONTEXT_SELECTION) {
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        item = gtk_menu_item_new_with_mnemonic (_("_Copy Text"));
+        g_signal_connect_swapped (item, "activate",
+                                  G_CALLBACK (webkit_web_view_copy_clipboard), view);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+    }
+
+    g_object_unref (result);
+    gdk_event_free (event);
+    gtk_widget_show_all (GTK_WIDGET (menu));
+}
+
 static gboolean
 view_navigation_requested (WebKitWebView             *view,
                            WebKitWebFrame            *frame,
@@ -537,13 +909,14 @@ view_navigation_requested (WebKitWebView             *view,
                            WebKitWebPolicyDecision   *decision,
                            gpointer                   user_data)
 {
+    const gchar *requri = webkit_network_request_get_uri (request);
     YelpViewPrivate *priv = GET_PRIV (view);
     YelpUri *uri;
 
-    debug_print (DB_FUNCTION, "entering\n");
-
-    uri = yelp_uri_new_relative (priv->uri,
-                                 webkit_network_request_get_uri (request));
+    if (g_str_has_prefix (requri, BOGUS_URI))
+        uri = yelp_uri_new_relative (priv->uri, requri + BOGUS_URI_LEN);
+    else
+        uri = yelp_uri_new_relative (priv->uri, requri);
 
     webkit_web_policy_decision_ignore (decision);
 
@@ -585,6 +958,12 @@ view_resource_request (WebKitWebView         *view,
     }
 }
 
+static void
+view_print (GtkAction *action, YelpView  *view)
+{
+    webkit_web_frame_print (webkit_web_view_get_main_frame (WEBKIT_WEB_VIEW (view)));
+}
+
 static void
 view_history_action (GtkAction *action,
                      YelpView  *view)
@@ -610,6 +989,8 @@ view_history_action (GtkAction *action,
 
     priv->back_load = TRUE;
     yelp_view_load_uri (view, ((YelpBackEntry *) priv->back_cur->data)->uri);
+    priv->vadjust = ((YelpBackEntry *) priv->back_cur->data)->vadj;
+    priv->hadjust = ((YelpBackEntry *) priv->back_cur->data)->hadj;
 }
 
 static void
@@ -842,12 +1223,13 @@ uri_resolved (YelpUri  *uri,
     GSList *proxies, *cur;
     GError *error;
     gchar *struri;
+    GParamSpec *spec;
 
     debug_print (DB_FUNCTION, "entering\n");
 
     switch (yelp_uri_get_document_type (uri)) {
     case YELP_URI_DOCUMENT_TYPE_EXTERNAL:
-        g_signal_emit (view, signals[EXTERNAL_URI], 0, uri, 0);
+        g_signal_emit (view, signals[EXTERNAL_URI], 0, uri);
         return;
     case YELP_URI_DOCUMENT_TYPE_NOT_FOUND:
         struri = yelp_uri_get_canonical_uri (uri);
@@ -942,11 +1324,15 @@ uri_resolved (YelpUri  *uri,
             gtk_widget_set_tooltip_text (GTK_WIDGET (cur->data), "");
     }
 
-    g_signal_emit_by_name (view, "notify::yelp-uri", 0);
+    spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                         "yelp-uri");
+    g_signal_emit_by_name (view, "notify::yelp-uri", spec);
 
     g_free (priv->page_id);
     priv->page_id = yelp_uri_get_page_id (priv->uri);
-    g_signal_emit_by_name (view, "notify::page-id", 0);
+    spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                         "page-id");
+    g_signal_emit_by_name (view, "notify::page-id", spec);
 
     view_load_page (view);
 }
@@ -965,15 +1351,19 @@ document_callback (YelpDocument       *document,
         gchar *prev_id, *next_id, *real_id;
         GtkAction *action;
         YelpBackEntry *back = NULL;
+        GParamSpec *spec;
 
         real_id = yelp_document_get_page_id (document, priv->page_id);
-        if (priv->page_id && g_str_equal (real_id, priv->page_id)) {
+        if (priv->page_id && real_id && g_str_equal (real_id, priv->page_id)) {
             g_free (real_id);
         }
         else {
+            GParamSpec *spec;
             g_free (priv->page_id);
             priv->page_id = real_id;
-            g_signal_emit_by_name (view, "notify::page-id", 0);
+            spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                                 "page-id");
+            g_signal_emit_by_name (view, "notify::page-id", spec);
         }
 
         g_free (priv->root_title);
@@ -1005,14 +1395,26 @@ document_callback (YelpDocument       *document,
         gtk_action_set_sensitive (action, next_id != NULL);
         g_free (next_id);
 
-        g_signal_emit_by_name (view, "notify::root-title", 0);
-        g_signal_emit_by_name (view, "notify::page-title", 0);
-        g_signal_emit_by_name (view, "notify::page-desc", 0);
-        g_signal_emit_by_name (view, "notify::page-icon", 0);
+        spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                             "root-title");
+        g_signal_emit_by_name (view, "notify::root-title", spec);
+
+        spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                             "page-title");
+        g_signal_emit_by_name (view, "notify::page-title", spec);
+
+        spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                             "page-desc");
+        g_signal_emit_by_name (view, "notify::page-desc", spec);
+
+        spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                             "page-icon");
+        g_signal_emit_by_name (view, "notify::page-icon", spec);
     }
     else if (signal == YELP_DOCUMENT_SIGNAL_CONTENTS) {
+        YelpUriDocumentType doctype;
 	const gchar *contents;
-        gchar *mime_type, *page_id, *frag_id;
+        gchar *mime_type, *page_id, *frag_id, *full_uri;
         page_id = yelp_uri_get_page_id (priv->uri);
         debug_print (DB_ARG, "    document.uri.page_id=\"%s\"\n", page_id);
         mime_type = yelp_document_get_mime_type (document, page_id);
@@ -1042,10 +1444,21 @@ document_callback (YelpDocument       *document,
            based on actual file locations, but in fact it doesn't matter.  So
            we just make a bogus URI that's easy to process later.
          */
-        if (frag_id != NULL)
-            priv->bogus_uri = g_strdup_printf ("%s%p#%s", BOGUS_URI, priv->uri, frag_id);
-        else
-            priv->bogus_uri = g_strdup_printf ("%s%p", BOGUS_URI, priv->uri);
+        doctype = yelp_uri_get_document_type (priv->uri);
+        full_uri = yelp_uri_get_canonical_uri (priv->uri);
+        if (g_str_has_prefix (full_uri, "file:/") &&
+            (doctype == YELP_URI_DOCUMENT_TYPE_TEXT ||
+             doctype == YELP_URI_DOCUMENT_TYPE_HTML ||
+             doctype == YELP_URI_DOCUMENT_TYPE_XHTML )) {
+            priv->bogus_uri = full_uri;
+        }
+        else {
+            g_free (full_uri);
+            if (frag_id != NULL)
+                priv->bogus_uri = g_strdup_printf ("%s%p#%s", BOGUS_URI, priv->uri, frag_id);
+            else
+                priv->bogus_uri = g_strdup_printf ("%s%p", BOGUS_URI, priv->uri);
+        }
         g_signal_handler_block (view, priv->navigation_requested);
         webkit_web_view_load_string (WEBKIT_WEB_VIEW (view),
                                      contents,
@@ -1054,6 +1467,44 @@ document_callback (YelpDocument       *document,
                                      priv->bogus_uri);
         g_signal_handler_unblock (view, priv->navigation_requested);
         g_object_set (view, "state", YELP_VIEW_STATE_LOADED, NULL);
+
+        /* If we need to set the GtkAdjustment or trigger the page title
+         * from what WebKit thinks it is (see comment below), we need to
+         * let the main loop run through.
+         */
+        if (priv->vadjust > 0 || priv->hadjust > 0 || priv->page_title == NULL)
+            while (g_main_context_pending (NULL))
+                g_main_context_iteration (NULL, FALSE);
+
+        /* Setting adjustments only work after the page is loaded. These
+         * are set by view_history_action, and they're reset to 0 after
+         * each load here.
+         */
+        if (priv->vadjust > 0) {
+            if (priv->vadjustment)
+                gtk_adjustment_set_value (priv->vadjustment, priv->vadjust);
+            priv->vadjust = 0;
+        }
+        if (priv->hadjust > 0) {
+            if (priv->hadjustment)
+                gtk_adjustment_set_value (priv->hadjustment, priv->hadjust);
+            priv->hadjust = 0;
+        }
+
+        /* If the document didn't give us a page title, get it from WebKit.
+         * We let the main loop run through so that WebKit gets the title
+         * set so that we can send notify::page-title before loaded. It
+         * simplifies things if YelpView consumers can assume the title
+         * is set before loaded is triggered.
+         */
+        if (priv->page_title == NULL) {
+            GParamSpec *spec;
+            priv->page_title = g_strdup (webkit_web_view_get_title (WEBKIT_WEB_VIEW (view)));
+            spec = g_object_class_find_property ((GObjectClass *) YELP_VIEW_GET_CLASS (view),
+                                                 "page-title");
+            g_signal_emit_by_name (view, "notify::page-title", spec);
+        }
+
         g_free (frag_id);
         g_free (page_id);
         g_free (mime_type);
